# Configuration File
_DEFAULT_SESSION = None  # Used to reuse generated session. DON'T CHANGE IT

USER_AGENT = "Barcelona 289.0.0.77.109 Android (31/12; 480dpi; 1080x2278; vivo; vivo 1939; 2004; qcom; en_US; 383877305)"

# request timeout - in seconds
TIMEOUT = 5

# Example {"http":"proxy_here","https":"proxy_here"} Accepts python dictionary.
PROXY = None

if __name__ == "__main__":
    pass



import httpx
import asyncio
import bs4
import json
from . import util
from . import config
from .constants import Path


def make_request(url=None, method=None, params=None, request_payload=None, session=None, timeout=None, **kwargs):
    global data_container
    # fmt: off - Turns off formatting for this block of code. Just for the readability purpose.
    def make_regular_request(request_payload):
        return validate_response(session.request(**request_payload))

    async def make_async_request(request_payload):
        pagination_data = request_payload.pop("pagination_data",None)
        if not pagination_data:
            response = await session.request(**request_payload)
            return validate_response(response)
        return await _handle_pagination(request_payload=request_payload,session=session,**{"end_cursor":pagination_data})

    async def make_concurrent_requests(request_payload):
        if isinstance(request_payload,list):
            tasks_list = [asyncio.create_task(make_async_request(payload)) for payload in request_payload]
        else:
            query_params = request_payload.pop("params",None)
            if not isinstance(query_params,list):
                return await make_async_request({"params":query_params} | request_payload) 
            if not query_params:
                return await make_async_request(request_payload)  
            tasks_list = [asyncio.create_task(make_async_request({"params":query} | request_payload)) for query in query_params]
        return await asyncio.gather(*tasks_list, return_exceptions=True)

    data_container = []
    method = method or "GET"
    timeout = timeout or config.TIMEOUT or 10
    proxies = config.PROXY or None
    ssl_verify = False if proxies else True
    concurrent_requests = False
    session = session or config._DEFAULT_SESSION or httpx.Client(follow_redirects=True, timeout=timeout, proxies=proxies, verify=ssl_verify)
    if request_payload:
        concurrent_requests = True if isinstance(request_payload,list) or (isinstance(request_payload,dict) and isinstance(request_payload.get("params"),list) or request_payload.get("pagination_data")) else False
        if concurrent_requests:
            connection_limits = httpx.Limits(max_connections=100, max_keepalive_connections=10, keepalive_expiry=5)
            headers,cookies = session.headers,session.cookies
            session = httpx.AsyncClient(limits=connection_limits,headers=headers,cookies=cookies,follow_redirects=True,timeout=timeout,proxies=proxies,verify=ssl_verify)
            try:
                return asyncio.run(make_concurrent_requests(request_payload))
            except KeyboardInterrupt:
                print("Interuppted..")
                return data_container
            except Exception as error:
                print(error)
                return data_container
    else:
        request_payload = {"method":method,"url":url,"params":params} | kwargs
    return make_regular_request(request_payload)


def generate_request_data(endpoint, placeholder=None, params=None, additional_payload=None, pagination=None, return_payload=False, **kwargs):
        # fmt: off - Turns off formatting for this block of code. Just for the readability purpose.
    method = kwargs.pop("method",None) or "GET"
    params = params or {}
    url = kwargs.pop("url",None) or Path.API_URL
    url = util.generate_url(domain=url, url_path=endpoint)
    params,data,json_data = params,kwargs.pop("data",None),kwargs.pop("json",None)
    if additional_payload and isinstance(additional_payload,dict):
        existing_payload = params or data or json_data
        key,values = additional_payload.popitem()
        values = values[0] if isinstance(values,list) and len(values) == 1 else values
        if isinstance(values,list):
            additional_payload = [json.dumps(existing_payload | {key:str(each_item)}) if params else existing_payload | {key:str(each_item)} for each_item in values]
        else:
            additional_payload = json.dumps(existing_payload | {key:str(values)}) if params else existing_payload | {key:str(values)}
    else:
        additional_payload = json.dumps(params) if params else data or json_data if any([data,json_data]) else {}
    data_key = "params" if params else "data" if data else "json" if json_data else None
    additional_payload = {data_key:additional_payload} if data_key and additional_payload else {}
    if placeholder and isinstance(placeholder,list):
        request_payload = [{"method": method, "url": url.format(each_item)} | additional_payload | kwargs for each_item in placeholder]
    else:
        request_payload = {"method": method, "url": url.format(placeholder)} | additional_payload | kwargs
    # fmt: on   
    if data and "data" not in request_payload.keys():
        request_payload.update({"data": data})
    if return_payload:
        return request_payload
    if pagination:
        request_payload.update({"pagination_data": pagination}) if isinstance(request_payload, dict) else [
            item.update({"pagination_data": pagination}) for item in request_payload]
    return make_request(request_payload=request_payload)


async def _handle_pagination(url=None, request_payload=None, session=None, end_cursor=None, **kwargs):
    global data_container
        # fmt: off  - Turns off formatting for this block of code. Just for the readability purpose.
    data_placeholder = {"data": [],"cursor_endpoint": None, "has_next_page": True}
    request_payload = request_payload or {"url": url} | kwargs
    cursor_key,end_cursor = end_cursor.popitem()
    while data_placeholder["has_next_page"]:
        try:
            if end_cursor:
                query_params = request_payload["params"] or json.dumps({cursor_key:""})
                query_params = json.loads(query_params)
                query_params[cursor_key] = end_cursor
                request_payload["params"] = json.dumps(query_params)
            response = await session.request(**request_payload)
            response = validate_response(response)
            end_cursor = util.find_nested_key(response,"next_max_id") or util.find_nested_key(response,"paging_tokens")
            end_cursor = end_cursor[0] if isinstance(end_cursor[0],str) else end_cursor[0].get("downwards",None) if isinstance(end_cursor[0],dict) else None
            more_threads =  util.find_nested_key(response,"downwards_thread_will_continue")
            more_threads = more_threads[0] if more_threads else True
            data_placeholder['data'].append(response)
            print(f"Page: {len(data_placeholder['data'])}", end="\r")
            if end_cursor:
                data_placeholder['cursor_endpoint'] = end_cursor
            else:
                data_placeholder["has_next_page"] = False

            if not data_placeholder["has_next_page"] or not more_threads:
                data_container.append(data_placeholder)
                return data_placeholder
        # fmt: on 
        except ConnectionError as error:
            print(error)
            continue

        except Exception as error:
            print(error)
            data_container.append(data_placeholder)
            return data_placeholder


def validate_response(response):
    try:
        response_text = ""
        soup = bs4.BeautifulSoup(response.content, "lxml")
        if "json" in response.headers["Content-Type"]:
            return util.check_for_errors(response.json())
        response_text = "\n".join(
            [line.strip() for line in soup.text.split("\n") if line.strip()])
        response.raise_for_status()
        return soup
    except Exception as error:
        # print(f"{error}\n\n{response_text}\n")
        raise error


if __name__ == '__main__':
    pass



import httpx
import json
import datetime
from . import util
from . import config
from .request_util import generate_request_data
from .constants import Setting, Path


class MetaThreads:
    def __init__(self):
        self.session = httpx.Client(follow_redirects=True)
        self.session.headers.update(util.generate_headers())
        self.logged_in_user = None

    @property
    def session(self):
        return self._session

    @session.setter
    def session(self, session):
        self._session = session
        config._DEFAULT_SESSION = session

    @property
    def me(self):
        return self.logged_in_user

    @property
    def user_id(self):
        return self.logged_in_user.get('pk', '')

    def login(self, username, password):
        """Log in with instagram credentials.

        Args:
            username (str): Instagram username/email.
            password (str/int): Instagram password.

        Returns:
            dict: Token data.
        """
        timestamp = int(datetime.datetime.now().timestamp())
        password = f'#PWD_INSTAGRAM_BROWSER:0:{timestamp}:{password}'
        payload = {"client_input_params": {"password": password, "contact_point": username, "device_id": Setting.ANDROID_ID},
                   "server_params": {"credential_type": "password", 'device_id': Setting.ANDROID_ID},
                   'bk_client_context': {"bloks_version": Setting.BLOK_VERSION_ID, "styles_id": "instagram"},
                   'bloks_versioning_id': Setting.BLOK_VERSION_ID}

        payload = {keys: json.dumps(values) if isinstance(
            values, dict) else values for keys, values in payload.items()}

        response = generate_request_data(
            Path.LOGIN_URL, method="POST", data=payload)
        token_data, self.logged_in_user = util.get_auth_token(response)
        return token_data

    def thread_id_decorator(original_function):
        def wrapper(self, *args, **kwargs):
            args = list(args)
            args[0] = self.get_thread_id(args[0])
            args = tuple(args)
            return original_function(self, *args, **kwargs)
        return wrapper

    def user_id_decorator(original_function):
        def wrapper(self, *args, **kwargs):
            args = list(args)
            args[0] = self.get_user_id(args[0])
            args = tuple(args)
            return original_function(self, *args, **kwargs)
        return wrapper

    def get_thread_id(self, thread_url):
        if not isinstance(thread_url, list):
            thread_url = [thread_url]
        thread_ids = [*filter(lambda thread: thread if isinstance(
            thread, int) or thread.isnumeric() else None, thread_url)]
        if len(thread_ids) != len(thread_url):
            pending_ids = [
                thread for thread in thread_url if thread not in thread_ids]
            new_thread_ids = [*map(util.shortcode_to_id, pending_ids)]
            thread_ids = [*thread_ids, *new_thread_ids]
        return thread_ids if len(thread_ids) > 1 else thread_ids[0] if thread_ids else None

    @thread_id_decorator
    def get_thread(self, thread_id):
        """Get thread details.

        Args:
            thread_id (str/int): Thread ID or URL.

        Returns:
            dict: Thread details.
        """
        return generate_request_data(Path.THREAD_ENDPOINT, thread_id)

    @thread_id_decorator
    def get_thread_replies(self, thread_id, cursor=None):
        """Get thread replies.

        Args:
            thread_id (str/int): Thread ID or URL.
            cursor (str, optional): Last endcursor point. (To start from where you left off the last time). Defaults to None.

        Returns:
            dict: Thread replies dataset. i.e. people who replied to the thread and replied content.
        """
        pagination_data = {"paging_token'": cursor, "count": 100}
        return generate_request_data(Path.THREAD_ENDPOINT, thread_id, pagination=pagination_data)

    @thread_id_decorator
    def get_thread_likes(self, thread_id):
        """Get thread likes.

        Args:
            thread_id (str/int): Thread ID or URL.

        Returns:
            dict: Dataset of users who liked the thread.
        """
        return generate_request_data(Path.THREAD_LIKERS, thread_id)

    def get_user_id(self, username):
        if not isinstance(username, list):
            username = [username]
        user_ids = [*filter(lambda user: user if isinstance(
            user, int) or user.isnumeric() else None, username)]
        if len(user_ids) != len(username):
            pending_users = [user for user in username if user not in user_ids]
            response = generate_request_data(
                Path.USER_PROFILE_ENDPOINT, pending_users)
            user_ids = [*user_ids, *[user['data']['user']['id']
                                     for user in response]]
        return user_ids if len(user_ids) > 1 else user_ids[0] if user_ids else None

    @user_id_decorator
    def get_user(self, user_id):
        """Get user information.

        Args:
            user_id (str/int): User ID or username.

        Returns:
            dict: User profile data.
        """
        params = {'is_prefetch': False, 'entry_point': 'profile',
                  'from_module': 'ProfileViewModel'}
        return generate_request_data(Path.USER_INFO_ENDPOINT, user_id, params=params)

    @user_id_decorator
    def get_user_threads(self, user_id, cursor=None):
        """Get threads posted by the user.

        Args:
            user_id (str/int): User ID or username.
            cursor (str, optional): Last endcursor point. (To start from where you left off the last time). Defaults to None.

        Returns:
            dict: Dataset of a user's posted threads from the user profile.
        """
        params = {"count": 100}
        pagination_data = {"max_id": cursor}
        return generate_request_data(Path.USER_THREAD_ENDPOINT, user_id, params=params, pagination=pagination_data)

    @user_id_decorator
    def get_user_threads_replies(self, user_id, cursor=None):
        """Get threads the user replied to.

        Args:
            user_id (str/int): User ID or username.
            cursor (str, optional): Last endcursor point. (To start from where you left off the last time). Defaults to None.

        Returns:
            dict: Dataset of threads the user replied to.
        """
        params = {"count": 100}
        pagination_data = {"max_id": cursor}
        return generate_request_data(Path.USER_THREAD_REPLIES_ENDPOINT, user_id, params=params, pagination=pagination_data)

    @user_id_decorator
    def get_user_friends(self, user_id, followers=False, following=False, cursor=None):
        """Get user followers/followings data

        Args:
            user_id (str/int): User ID or username.
            followers (bool, optional): Set to True, if want to get a list of user followers. Defaults to False.
            following (bool, optional): Set to True, if want to get a list of user following. Defaults to False.
            cursor (str, optional): Last endcursor point. (To start from where you left off the last time). Defaults to None.
        Raises:
            Exception: Set one of the (followers,following) to True.

        Returns:
            dict: User friends (followers/following) dataset.
        """
        if (not followers and not following) or (followers and following):
            raise Exception("Set one of the (followers,following) to True.")

        endpoint = Path.USER_FOLLOWERS_ENDPOINT if followers else Path.USER_FOLLOWING_ENDPOINT
        params = {'count': 100, 'include_user_count': True,
                  'search_surface': 'barcelona_following_graph_page'}
        pagination_data = {"max_id": cursor}
        return generate_request_data(endpoint, user_id, params=params, pagination=pagination_data)

    def search_user(self, search_query):
        """Perform a user search

        Args:
            search_query (str): Search query you want to perform. i.e. username/full name.

        Returns:
            dict: Search results.
        """
        params = {'search_surface': 'user_search_page',
                  'timezone_offset': 0, 'count': 100}
        additional_payload = {'q': search_query}
        return generate_request_data(Path.SEARCH_USER_ENDPOINT, params=params, additional_payload=additional_payload)

    def get_notifications(self):
        """Get notification informations

        Returns:
            dict: Threads notifications.
        """
        params = {'feed_type': 'all', 'mark_as_seen': False,
                  'timezone_offset': 0, 'timezone_name': 'GMT'}
        return generate_request_data(Path.NOTIFICATIONS, params=params)

    # def send_inbox_seen(self):
    #     data = {"_uuid": Setting.DEVICE_ID}
    #     return generate_request_data(Path.INBOX, data=data, method="POST")

    @thread_id_decorator
    def repost_thread(self, thread_id):
        """Repost a thread

        Args:
            thread_id (str/int): Thread ID or URL.

        Returns:
            dict: Reposted thread data.
        """
        data = {"_uuid": Setting.DEVICE_ID}
        additional_payload = {'media_id': thread_id}
        return generate_request_data(Path.REPOST_THREAD, data=data, additional_payload=additional_payload, method="POST")

    @thread_id_decorator
    def delete_repost(self, thread_id):
        """Delete/Destroy a reposted thread.

        Args:
            thread_id (str/int): Thread ID or URL.

        Returns:
            dict: Destroyed thread data.
        """
        data = {'_uuid': Setting.DEVICE_ID}
        additional_payload = {'original_media_id': thread_id}
        return generate_request_data(Path.DELETE_REPOST, data=data, additional_payload=additional_payload, method="POST")

    def post_thread(self, thread_caption, reply_control="all"):
        """Post/Upload a new thread.

        Args:
            thread_caption (str): Thread caption/content to be posted.
            reply_control (str, optional): Choose who can reply to your thread. Defaults to "all". Available args ("all","followers","mentions")

        Returns:
            dict: Dataset of a newly posted thread.
        """
        reply_control = reply_control.lower()
        reply_control = 0 if reply_control == 'all' else 1 if reply_control == 'followers' else 2 if reply_control == 'mentions' else 0
        upload_id = int(datetime.datetime.now().microsecond *
                        datetime.datetime.now().microsecond)
        data = {"publish_mode": "text_post", "upload_id": upload_id, "text_post_app_info": {"reply_control": reply_control},
                "timezone_offset": 0, "_uid": self.user_id, "device_id": Setting.ANDROID_ID,
                "_uuid": Setting.DEVICE_ID, "caption": thread_caption, "audience": "default"}
        signed_data = {"signed_body": f"SIGNATURE.{json.dumps(data)}"}
        return generate_request_data(Path.POST_THREAD, data=signed_data, method="POST")

    def delete_thread(self, thread_id):
        """Delete/Destroy a thread posted by the user. Note : Doesn't work with a reposted thread. Check delete_repost for that case.

        Args:
            thread_id (str/int): Thread ID or URL.

        Returns:
            dict: Dataset of a deleted thread.
        """
        if isinstance(thread_id, int) or isinstance(thread_id, str):
            if str(thread_id).find("_") == -1:
                thread_id = f"{thread_id}_{self.user_id}"
        if isinstance(thread_id, list):
            thread_id = [f"{item}_{self.user_id}" if str(thread_id).find(
                "_") == -1 else item for item in thread_id]
        params = {'media_type': 'TEXT_POST'}
        data = {"media_id": thread_id,
                "_uid": self.user_id, "_uuid": Setting.DEVICE_ID}
        signed_data = {"signed_body": f"SIGNATURE.{json.dumps(data)}"}
        return generate_request_data(Path.DELETE_THREAD, thread_id, params=params, data=signed_data, method="POST")

    @thread_id_decorator
    def like_thread(self, thread_id):
        """Like a thread.

        Args:
            thread_id (str/int): Thread ID or URL.

        Returns:
            dict: Liked thread data.
        """
        data = {"delivery_class": "organic", "tap_source": "button",
                "media_id": thread_id, "_uid": self.user_id, "_uuid": Setting.DEVICE_ID}
        signed_data = {"signed_body": f'SIGNATURE.{data}', "d": 0}
        return generate_request_data(Path.LIKE_THREAD, thread_id, data=signed_data, method="POST")

    @thread_id_decorator
    def unlike_thread(self, thread_id):
        """Unlike a thread.

        Args:
            thread_id (str/int): Thread ID or URL.

        Returns:
            dict: Unliked thread data.
        """
        data = {"delivery_class": "organic", "tap_source": "button",
                "media_id": thread_id, "_uid": self.user_id, "_uuid": Setting.DEVICE_ID}
        signed_data = {"signed_body": f'SIGNATURE.{data}', "d": 0}
        return generate_request_data(Path.UNLIKE_THREAD, thread_id, data=signed_data, method="POST")

    @user_id_decorator
    def follow(self, user_id):
        """Follow a user.

        Args:
            user_id (str/int): User ID or username.

        Returns:
            dict: Followed user data.
        """
        data = {"user_id": user_id, "_uid": self.user_id,
                "_uuid": Setting.DEVICE_ID}
        signed_data = {"signed_body": f'SIGNATURE.{data}', "d": 0}
        return generate_request_data(Path.FOLLOW_USER, user_id, data=signed_data, method="POST")

    @user_id_decorator
    def unfollow(self, user_id):
        """Unfollow a user.

        Args:
            user_id (str/int): User ID or username.

        Returns:
            dict: Unfollowed user data.
        """
        data = {"user_id": user_id, "_uid": self.user_id,
                "_uuid": Setting.DEVICE_ID}
        signed_data = {"signed_body": f'SIGNATURE.{data}', "d": 0}
        return generate_request_data(Path.UNFOLLOW_USER, user_id, data=signed_data, method="POST")


if __name__ == "__main__":
    pass




class Setting:
    BLOK_VERSION_ID = "5f56efad68e1edec7801f630b5c122704ec5378adbee6609a448f105f34a9c73"
    ANDROID_ID = ""
    DEVICE_ID = ""


class Path:
    HOST = "i.instagram.com"
    API_URL = "https://i.instagram.com/api/v1/"
    LOGIN_URL = "bloks/apps/com.bloks.www.bloks.caa.login.async.send_login_request/"
    # endpoints
    USER_PROFILE_ENDPOINT = "https://www.instagram.com/api/v1/users/web_profile_info/?username={}"
    THREAD_ENDPOINT = "text_feed/{}/replies/"
    THREAD_LIKERS = "media/{}/likers/"
    USER_THREAD_ENDPOINT = "text_feed/{}/profile/"
    USER_THREAD_REPLIES_ENDPOINT = "text_feed/{}/profile/replies/"
    USER_INFO_ENDPOINT = "users/{}/info"
    SEARCH_USER_ENDPOINT = "users/search/"
    USER_FOLLOWERS_ENDPOINT = "friendships/{}/followers/"
    USER_FOLLOWING_ENDPOINT = "friendships/{}/following/"
    # actions
    FOLLOW_USER = "friendships/create/{}/"
    UNFOLLOW_USER = "friendships/destroy/{}/"
    NOTIFICATIONS = "text_feed/text_app_notifications/"
    INBOX = "text_feed/text_app_inbox_seen/"
    LIKE_THREAD = "media/{}/like/"
    UNLIKE_THREAD = "media/{}/unlike/"
    REPOST_THREAD = "repost/create_repost/"
    DELETE_REPOST = "repost/delete_text_app_repost/"
    POST_THREAD = "media/configure_text_only_post/"
    DELETE_THREAD = "media/{}/delete/"


if __name__ == "__main__":
    pass
